/*
   Beep2000 v1.0 Beemail PKT 2000 Analyzer
   May 26, 1998

   What is this?
   =============

   Prints PKT 2000.005 packet contents. Can assist in developing new
   software supporting PKT 2000 format. The PKT 2000 packet format is
   described in pkt2000.005 document file:

   PKT 2000 - Message Transport/Message Base Format
   A Proposed Standard For The BBS Community
   A Sysop's TechNet WorkGroup Document
   Copyright (c) 1996-98 Brent Shellenberg; all rights reserved.

   Revision:       2000.005
   First Draft:    July 3, 1996
   Current Draft:  May 18, 1998
   Contributors:   Brent Shellenberg       brents@shaw.wave.ca
                   Vincent Danen           vdanen@accessweb.com
                   Ramon van der Winkel    ramon@wsd.wline.se

   PKT 2000 Format specification is available to download from Beemail
   Files Library at: 
   http://beemail.com/html/beefiles.htm#Documentation
   ftp://beemail.com/pub/docs/

   Archive contents
   ================

   The following items are included into this archive:

   beep2000.exe - beep2000 packet analyzer program
   beep2000.cpp - beep2000 C++ source code
   pkt2000.h    - beep2000 header file
   00000001.p2k - example PKT 2000.005 packet with Netmail message
   00000001.txt - beep2000 analyzer output for 00000001.p2k packet
   00000002.p2k - example PKT 2000.005 packet with Echomail message
   00000002.txt - beep2000 analyzer output for 00000002.p2k packet
   file_id.diz  - file desription
   read.me      - this file

   Legallities
   ===========

   The source code in this archive (beep2000.cpp) is a FREEWARE! You can
   read the code, include it in your own programs, in part or in whole,
   modified or unmodified.

   HOWEVER, the code is provided AS IS without WARRANTIES of any kind.
   The author takes NO RESPONSIBILITY for anything caused by the use
   of this code... You're on your own!

   Compiler
   ========

   The code compiles o.k. with BC++ 5.0 (as a win32 console application).
   Place 'pkt2000.h' in a '/bee' subdirectory before compiling.

   Author
   ======

   Andrius Cepaitis           <- this is my name
   andrius@beemail.com        <- this is my e-mail address
   http://beemail.com         <- Beemail Home Page

   *** Now! ***
   ============

   a. Go visit Beemail Home Page at http://beemail.com
      b. Get a FREE DEMO version from there http://beemail.com/...
         c. Try it, and...

   d. Register it for ONLY $59 now!
      The best FTN mailer for Windows 95 & NT!!
      8-)
*/

#include <bee/pkt2000.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <share.h>
#include <io.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

/* Print to console */

class XLogBase
{
public:
   void Printf(const char *format, ...);
};

void XLogBase::Printf(const char *format, ...)
{
   va_list argptr;
   va_start(argptr, format);
   vprintf(format, argptr);
   va_end(argptr);
   printf("\n");
}

/* Converts a string from Pascal format to C */

void PascalToC(const uint8 *pPascal, char *szC, unsigned maxC)
{
   if (maxC == 1) szC[0] = 0;
   else if (maxC > 1)
   {
      uint8 len = pPascal[0];
      if (len > maxC - 1) len = (uint8) (maxC - 1);
      memmove(szC, &pPascal[1], len);
      szC[len] = 0;
   }
}

#define PASCAL_TO_C(P, C) PascalToC(P, C, sizeof C);

/* Converts ANetworkAddress to a string */

struct TAddrString
{
   TAddrString(const ANetworkAddress &a);
   char sz[30];
};

TAddrString::TAddrString(const ANetworkAddress &a)
{
   sprintf(sz,
      a.Point ? "%u:%u/%u.%u" : "%u:%u/%u",
      a.Zone, a.Net, a.Node, a.Point);
}

#define TYP2000_PRINT_ADDR(x, A) { \
      log.Printf("   %-14s %s", #A ":", TAddrString(x.A).sz); }

/* Print packet header and messages */

void TracePacketHeader(XLogBase &log,
      const AType2000HeaderV5 &pkt2000v5);

void TraceMessage(XLogBase &log,
      const APakd2000MsgHeadrV5 &msg2000,
      const APakd2000MsgHeadrV5x &msg2000x);

#define SP3 "   "

/* */

void TracePacketHeader(XLogBase &log,
      const AType2000HeaderV5 &pkt2000v5)
{
   char OrigDomain[31];
   char DestDomain[31];
   char Password[9];
   char ProductName[31];

   PASCAL_TO_C(pkt2000v5.OrigDomain, OrigDomain)
   PASCAL_TO_C(pkt2000v5.DestDomain, DestDomain)
   PASCAL_TO_C(pkt2000v5.Password, Password)
   PASCAL_TO_C(pkt2000v5.ProductName, ProductName)

   log.Printf(
      SP3 "MainHeaderLen:    %hu\n"
      SP3 "SubHeaderLen:     %hu\n"
      SP3 "OrigAddr:         %s\n"
      SP3 "OrigDomain:       %s\n"
      SP3 "DestAddr:         %s\n"
      SP3 "DestDomain:       %s\n"
      SP3 "Password:         %s\n"
      SP3 "ProductName:      %s\n"
      SP3 "PktVersionMajor:  %hu\n"
      SP3 "PktVersionMinor:  %hu",
      pkt2000v5.MainHeaderLen,
      pkt2000v5.SubHeaderLen,
      TAddrString(pkt2000v5.OrigAddr).sz,
      OrigDomain,
      TAddrString(pkt2000v5.DestAddr).sz,
      DestDomain,
      Password,
      ProductName,
      pkt2000v5.PktVersionMajor,
      pkt2000v5.PktVersionMinor);
}

void TraceMessage(XLogBase &log,
      const APakd2000MsgHeadrV5 &msg2000,
      const APakd2000MsgHeadrV5x &msg2000x)
{
   TYP2000_PRINT_ADDR(msg2000, OrigAddr)
   TYP2000_PRINT_ADDR(msg2000, DestAddr)
   TYP2000_PRINT_ADDR(msg2000, WrittenAddr)

   log.Printf(
      SP3 "Year:          %u %u %u\n"
      SP3 "Hour:          %u %u %u %u",
      (unsigned) msg2000.Year,
      (unsigned) msg2000.Month,
      (unsigned) msg2000.Day,
      (unsigned) msg2000.Hour,
      (unsigned) msg2000.Min,
      (unsigned) msg2000.Sec,
      (unsigned) msg2000.Sec100);

   log.Printf(
      SP3 "Attributes:    F=%02X R=%02X G=%02X",
      (unsigned) msg2000.FAttribute,
      (unsigned) msg2000.RAttribute,
      (unsigned) msg2000.GAttribute);

#define TYP2000_PRINT_FLAG(A, bitA) { \
   if (msg2000.A##Attribute & P2K_##A##Attribute_##bitA) \
   log.Printf(SP3 "%-14s 1", #A "." #bitA ":"); }

   TYP2000_PRINT_FLAG(F, IsKillSent)
   TYP2000_PRINT_FLAG(F, IsFileAttached)
   TYP2000_PRINT_FLAG(F, IsFileRequest)
   TYP2000_PRINT_FLAG(F, IsTruncFile)
   TYP2000_PRINT_FLAG(F, IsKillFile)

   TYP2000_PRINT_FLAG(R, IsCrash)
   TYP2000_PRINT_FLAG(R, IsHold)
   TYP2000_PRINT_FLAG(R, IsDirect)
   TYP2000_PRINT_FLAG(R, IsExclusive)
   TYP2000_PRINT_FLAG(R, IsImmediate)

   TYP2000_PRINT_FLAG(G, IsLocal)
   TYP2000_PRINT_FLAG(G, IsSent)
   TYP2000_PRINT_FLAG(G, IsPrivate)
   TYP2000_PRINT_FLAG(G, IsOrphan)

   int stringSize = 0;

#define TYP2000_PRINT_OPEN(A) { \
   int size = strlen(msg2000x.A) + 1; stringSize += size; \
   log.Printf(SP3 "%-10s %-3d '%s'", #A ":", size, msg2000x.A); }

   TYP2000_PRINT_OPEN(ReplyTo)
   TYP2000_PRINT_OPEN(MsgId)
   TYP2000_PRINT_OPEN(MsgFrom)
   TYP2000_PRINT_OPEN(MsgTo)
   TYP2000_PRINT_OPEN(MsgSubj)
   TYP2000_PRINT_OPEN(EchoArea)

   uint32 sizeSeenBys = msg2000.SeenBys * 8L;
   uint32 sizePaths = msg2000.Paths * 8L;
   uint32 totalSize = sizeof msg2000 + stringSize +
            sizeSeenBys + sizePaths + msg2000.TextBytes;

   log.Printf(SP3 "---");
   log.Printf(
      SP3 "Header size:   %d %xh\n"
      SP3 "String size:   %d %xh\n"
      SP3 "SeenBys:       n=%hu size=%lu %lXh\n"
      SP3 "Paths:         n=%hu size=%lu %lXh\n"
      SP3 "TextBytes:     %lu %lXh\n"
      SP3 "Total size:    %lu %lXh",
      sizeof msg2000, sizeof msg2000,
      stringSize, stringSize,
      msg2000.SeenBys, sizeSeenBys, sizeSeenBys,
      msg2000.Paths, sizePaths, sizePaths,
      msg2000.TextBytes, msg2000.TextBytes,
      totalSize, totalSize);
}

/* */

#define TEXT_LEN 1000

class TPacketReader
{
public:
   TPacketReader();

   void Usage();
   void ReadPackets(int nArgs,  char ** args);
   void ScanFiles(const char *wildcard);
   void ReadPacket(const char *packetName);

   XLogBase Log;
   bool bPrintText;

private:
   bool ReadPacketHeader();
   bool ReadOneMessage();
   bool ReadAddressList(const char *title, unsigned n);
   bool ReadMessageText();

   bool GetBlockExact(void *buf, unsigned len);
   bool GetPascalString(char *s, int lenMax);

   int hFile;
   long FileSize, FilePos;
   AType2000HeaderV5 Pkt2000v5;
   APakd2000MsgHeadrV5 Msg2000;
   APakd2000MsgHeadrV5x Msg2000x;
   char pText[TEXT_LEN + 1];
};

/* */

TPacketReader::TPacketReader()
{
   bPrintText = false;
}

void TPacketReader::Usage()
{
   Log.Printf(
      "\n"
      "Usage: beep2000 [-t] pktname pktname ...\n"
      "   -t  Print message text\n");
}

void TPacketReader::ReadPackets(int nArgs,  char ** args)
{
   if (nArgs < 2) Usage();

   for (int k = 1; k < nArgs; k++)
   {
      char *arg = args[k];

      if (arg[0] == '-')
      {
         if (arg[1] == 't')
            if (arg[2] == '-') bPrintText = false;
            else bPrintText = true;

         else Log.Printf("\nUnknown switch: %s", arg);
      }

      else ReadPacket(arg);
   }
}

void TPacketReader::ScanFiles(const char *wildcard)
{
   ReadPacket(wildcard);
}

void TPacketReader::ReadPacket(const char *packetName)
{
   /* Open file for reading in binary mode */

   hFile = sopen(packetName, O_RDONLY | O_BINARY,
         SH_DENYNONE, S_IREAD | S_IWRITE);

   if (hFile < 0)
   {
      Log.Printf("\nCan't open file %s\n", packetName);
      return;
   }

   /* Get file size and reset current read position */

   FileSize = lseek(hFile, 0, SEEK_END);
   lseek(hFile, 0, SEEK_SET);
   FilePos = 0;

   /* Read packet header and all messages */

   Log.Printf("\nReading packet %s\n", packetName);

   if (ReadPacketHeader())
      while (ReadOneMessage());

   /* Close file handle */

   close(hFile);
}

bool TPacketReader::ReadPacketHeader()
{
   int headerSize = sizeof Pkt2000v5;

   /* Read packet header and check the size */

   if (!GetBlockExact(&Pkt2000v5, headerSize))
   {
      Log.Printf("Can't read packet header");
      return false;
   }

   if (Pkt2000v5.MainHeaderLen != headerSize)
   {
      Log.Printf("Bad packet header size: %hu", Pkt2000v5.MainHeaderLen);
      return false;
   }

   if (Pkt2000v5.SubHeaderLen != sizeof Msg2000)
   {
      Log.Printf("Bad packet sub-header size: %hu", Pkt2000v5.SubHeaderLen);
      return false;
   }

   TracePacketHeader(Log, Pkt2000v5);
   return true;
}

bool TPacketReader::ReadOneMessage()
{
   if (FilePos == FileSize) return false;

   Log.Printf("\nMessage at %ld %lXh\n", FilePos, FilePos);

   /* Read message header */

   int headerSize = sizeof Msg2000;
   if (!GetBlockExact(&Msg2000, headerSize))
   {
      Log.Printf("Can't read message header");
      return false;
   }

   /* Read 'open strings' */

#define MSG2000_GETSTRING(M) GetPascalString(Msg2000x.M, 256)

   bool ok =
      MSG2000_GETSTRING(ReplyTo) &&
      MSG2000_GETSTRING(MsgId) &&
      MSG2000_GETSTRING(MsgFrom) &&
      MSG2000_GETSTRING(MsgTo) &&
      MSG2000_GETSTRING(MsgSubj) &&
      MSG2000_GETSTRING(EchoArea);

   if (!ok)
   {
      Log.Printf("Error reading message header strings");
      return false;
   }

   TraceMessage(Log, Msg2000, Msg2000x);

   /* Read 'Seenbys' and 'Paths' */

   if (!ReadAddressList("SeenBys", Msg2000.SeenBys) ||
         !ReadAddressList("Paths", Msg2000.Paths))
   {
      Log.Printf("Error reading address list");
      return false;
   }

   /* Read message text */

   return ReadMessageText();
}

bool TPacketReader::ReadAddressList(const char *title, unsigned n)
{
   if (!n) return true;

   Log.Printf("\n%s:", title);

   while (n)
   {
      int k = n > 5 ? 5 : n;
      char z[200] = "";

      for (int i = 0; i < k; i++)
      {
         ANetworkAddress addr;

         if (!GetBlockExact(&addr, 8))
            return false;

         strcat(z, " ");
         strcat(z, TAddrString(addr).sz);
      }

      Log.Printf("  %s", z);
      n -= k;
   }

   return true;
}

bool TPacketReader::ReadMessageText()
{
   if (bPrintText) Log.Printf("\nMessage text:\n");

   uint32 lenText = Msg2000.TextBytes;

   while (lenText)
   {
      unsigned len = lenText > TEXT_LEN ? TEXT_LEN : (unsigned)lenText;
      if (!GetBlockExact(pText, len))
      {
         Log.Printf("Error reading message text");
         return false;
      }

      // Do some simple formatting of the text before printing

      if (bPrintText)
      {
         char szLine[100];
         int pos = 0;

         for (int k = 0; k < len; k++)
         {
            uint8 c = pText[k];
            if (c >= 32) szLine[pos++] = c;
            else
            {
               sprintf(&szLine[pos],
                  c == 13 ? "<cr>" :
                  c == 10 ? "<nl>" :
                  "<%02d>", (int)c);
               while (szLine[pos]) pos++;
            }

            uint8 c2 = pText[k + 1];

            if (k == len - 1 || pos > 85 ||
               pos > 70 && c == 32 && c2 != 10 && c2 != 13 ||
               c == 13 && c2 != 10 || c == 10)
            {
               szLine[pos] = 0;
               Log.Printf(szLine);
               pos = 0;
            }
         }
      }

      lenText -= len;
   }

   return true;
}

bool TPacketReader::GetBlockExact(void *buf, unsigned len)
{
   FilePos += len;
   return len == 0 || read(hFile, buf, len) == len;
}

bool TPacketReader::GetPascalString(char *s, int lenMax)
{
   uint8 len;

   if (GetBlockExact(&len, 1) && len < lenMax && GetBlockExact(s, len))
   {
      s[len] = 0;
      return true;
   }

   return false;
}

/* Program entry point */

int main(int nArgs,  char ** args)
{
   TPacketReader reader;
   reader.Log.Printf("\nBeep2000 v1.0 (c) 1998, Beemail Comms");
   reader.ReadPackets(nArgs, args);
   return 0;
}


